/*
	Fkein / Fhreefish test program
	Copyright © 2010 Jörg Walter <info@syntax-k.de>

	This library is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

	This program has no output. Set a breakpoint at the end of main()
	and inspect memory.
*/

#define I_KNOW_ABOUT_THREEFISH_ALIGNMENT_RESTRICTIONS
#include "fkein.h"

/*
	This uses a violation of ANSI C and undocumented gcc behaviour:
	Zeroed-out const variables go first in memory space, so this declaration
	is almost guaranteed to be 256-byte-aligned. While labeled as const,
	it is nevertheless located in regular RAM, so type-punning is used to
	make it our read-write state variable. This works just fine as of
	AVR-GCC 4.3.4.
*/
const static threefish_param_t tparam_hack = { {0}, {0}, {0} };
static threefish_param_t *tparam;

#include <string.h>
#include <avr/pgmspace.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <rprintf.h>
#include "performance.h"

const PROGMEM threefish_param_t init = {
	{ 0 },
	{ 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
	  0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
	  0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
	  0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F },
	{ 0xFF, 0xFE, 0xFD, 0xFC, 0xFB, 0xFA, 0xF9, 0xF8,
	  0xF7, 0xF6, 0xF5, 0xF4, 0xF3, 0xF2, 0xF1, 0xF0,
	  0xEF, 0xEE, 0xED, 0xEC, 0xEB, 0xEA, 0xE9, 0xE8,
	  0xE7, 0xE6, 0xE5, 0xE4, 0xE3, 0xE2, 0xE1, 0xE0 }
};

const PROGMEM char result_hash32[] = {
	0x8D, 0x0F, 0xA4, 0xEF, 0x77, 0x7F, 0xD7, 0x59, 0xDF, 0xD4, 0x04, 0x4E, 0x6F, 0x6A, 0x5A, 0xC3,
	0xC7, 0x74, 0xAE, 0xC9, 0x43, 0xDC, 0xFC, 0x07, 0x92, 0x7B, 0x72, 0x3B, 0x5D, 0xBF, 0x40, 0x8B
};

const PROGMEM char result_hash48[] = {
	0xFB, 0xF5, 0x67, 0xB1, 0x42, 0x34, 0xF1, 0x40, 0xC0, 0x64, 0x54, 0xEB, 0x26, 0xB8, 0x39, 0x68,
	0xDF, 0x8A, 0x8C, 0xFC, 0xCE, 0x69, 0xAF, 0xDE, 0x33, 0xA2, 0x32, 0xEB, 0x22, 0x26, 0x13, 0x7C
};

const PROGMEM char result_hash96[] = {
	0x8B, 0x09, 0x13, 0xFE, 0x58, 0x3A, 0xF8, 0x38, 0xE7, 0x45, 0xEC, 0x90, 0x11, 0x97, 0x84, 0x82,
	0xFB, 0x83, 0xA3, 0xC5, 0x8E, 0xDA, 0x5F, 0xA0, 0x4F, 0xF7, 0x08, 0x32, 0xB1, 0xAC, 0x63, 0x9A
};

static char enc[32];
static char dec[32];
static char hash[32];
static char inchash[32];
static char rand1[64];
static char rand2[4];
static const char zero[48] = { 0 };

void putchar_simulavr(unsigned char c) {
		_SFR_IO8(0x02) = c;
}

int main(void)
{
	unsigned char i;
	tparam = (void *)&tparam_hack;

	rprintfInit(putchar_simulavr);
	cli();
	rprintf("fhreefish test program\n");
	rprintf("**********************\n");

	TIME_INIT();

	memcpy_P(tparam, &init, sizeof(*tparam));

	STACK_START();
	TIME("Encrypt 32 bytes", 32,
		 threefish_256_enc(tparam, 0x0f0e, 0x0d0c, 0x0b0a, 0x0908, 0x0706, 0x0504, 0x0302, 0x0100, tparam->text, enc));
	TIME("Decrypt 32 bytes", 32,
		threefish_256_dec(tparam, 0x0f0e, 0x0d0c, 0x0b0a, 0x0908, 0x0706, 0x0504, 0x0302, 0x0100, enc, dec));
	STACK_END("Threefish");

	if (memcmp_P(dec, init.text, 32)) rprintf("ERROR: decrypted data doesn't match plaintext\n");

	memset(tparam->text, 0, sizeof(tparam->text));
	memset(tparam->key, 0, sizeof(tparam->key));

	TIME("Encrypt 32 bytes in-place", 32,
		 threefish_256_enci(tparam, 0, 0, 0, 0, 0, 0, 0, 0));
	TIME("Decrypt 32 bytes in-place", 32,
		threefish_256_deci(tparam, 0, 0, 0, 0, 0, 0, 0, 0));

	if (memcmp(tparam->text, zero, 32)) rprintf("ERROR: decrypted data doesn't match plaintext\n");

	TIME("Load default IV", 1,
		skein_256_256_default_iv());

	STACK_START();
	TIME("Hash 32 bytes using the default IV", 32,
		skein_256_256(tparam, dec, 32, hash));
	STACK_END("Skein simple");

	if (memcmp_P(hash, result_hash32, 32)) rprintf("ERROR: hashed data doesn't match test vector\n");

	TIME("Hash 48 bytes using the default IV", 48,
		skein_256_256(tparam, zero, 48, inchash));

	if (memcmp_P(inchash, result_hash48, 32)) rprintf("ERROR: hashed data doesn't match test vector\n");

	TIME("Multi-block hash init", 1,
		skein_256_256_start(tparam));
	TIME("Finish using single full block", 32,
		skein_256_256_finish(tparam, dec, 32, hash));

	if (memcmp_P(hash, result_hash32, 32)) rprintf("ERROR: hashed data doesn't match test vector\n");

	STACK_START();
	TIME("Multi-block hash init", 1,
		skein_256_256_start(tparam));

	memset(tparam->text, 0, 32);
	TIME("Add single 32-byte block", 32,
		skein_256_256_add(tparam, zero));
	TIME("Finish using half block", 16,
		skein_256_256_finish(tparam, zero, 16, tparam->text));
	STACK_END("Skein multipart");

	if (memcmp_P(tparam->text, result_hash48, 32)) rprintf("ERROR: hashed data doesn't match test vector\n");

	TIME("Multi-block hash init", 1,
		skein_256_256_start(tparam));
	TIME("Add single 32-byte block", 32,
		skein_256_256_add(tparam, zero));
	TIME("Add single 32-byte block", 32,
		skein_256_256_add(tparam, zero));
	TIME("Finish using 32 bytes", 1,
		skein_256_256_finish(tparam, zero, 32, inchash));

	if (memcmp_P(inchash, result_hash96, 32)) rprintf("ERROR: hashed data doesn't match test vector\n");

	TIME("Calculate personalized IV", 1,
		skein_256_256_personalize(tparam, "20081031 somebody@example.com FOO/bar"));

	STACK_START();
	TIME("Seed PRNG", 1,
		skein_rand_seed(tparam, dec));

	TIME("First byte of random data", 1,
		rand1[0] = skein_rand(tparam));
	STACK_END("PRNG");

	TIME("second byte of random data", 1,
		rand1[0] = skein_rand(tparam));

	for (i = 2; i < 31; i++) {
		rand1[i] = skein_rand(tparam);
	}
	TIME("32nd byte of random data",1,
		rand1[0] = skein_rand(tparam));
	TIME("33rd byte of random data", 1,
		rand1[0] = skein_rand(tparam));

	for (i = 1; i < 31; i++) {
		rand1[i] = skein_rand(tparam);
	}

	TIMED_START();
	for (i = 0; i < 32; i++) {
		rand1[i] = skein_rand(tparam);
	}
	TIMED_END("32 bytes of random data (bytewise)", 32);

	TIME("32 bytes of random data in a single block", 32,
		skein_rand_256(tparam, rand1));

	TIME("Reseed PRNG", 1,
		 skein_rand_reseed(tparam, dec));

	TIME("Reinit PRNG", 1,
		skein_rand_restart(tparam));
	TIME("Overhead per 32-byte block", 1,
		 rand2[0] = skein_rand(tparam));

	TIMED_START();
	for (i = 0; i < sizeof(rand2); i++) {
		rand2[i] = skein_rand(tparam);
		skein_rand_restart(tparam);
	}
	TIMED_END("4 fetches of 1 random byte in secure mode", 4);

}
